//  Copyright 2021 Aaron Hance
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

{
    Driver->Storage->VolumeManager - Volume manager for storage devices.

    @author(Aaron Hance ah@aaronhance.me)
}

unit volumemanager;

interface

uses
    console,
    drivermanagement,
    drivertypes,
    lists,
    lmemorymanager,
    MBR,
    rtc,
    serial,
    storagetypes,
    strings,
    terminal,
    tracer,
    util;

var
    storageDevices : PDList; //index in this array is global drive id
    volumes : PDList;

    procedure init();
    procedure register_device(device : PStorage_Device);
    function discover_volumes(device : PStorage_device) : PDList;

    function get_volume_list() : PDList;
    

implementation

procedure init();
begin
    push_trace('VolumeManager.init');

    storageDevices := DL_New(sizeof(PStorage_Device));
    volumes := DL_New(sizeof(PStorage_Volume));

    //TODO register commnads
    pop_trace();
end;

procedure register_device(device : PStorage_Device);
var 
    diskVolumes : PDList;
begin
    push_trace('VolumeManager.register_device');

    DL_Add(storageDevices);
    DL_Set(storageDevices, DL_Size(storageDevices) - 1, puint32(device));

    diskVolumes := discover_volumes(device);

    if volumes <> nil then begin
        volumes := DL_Concat(volumes, diskVolumes);
    end;

    pop_trace();
end;

function discover_volumes(device : PStorage_device) : PDList;
begin
    push_trace('VolumeManager.discover_volumes');
    
    discover_volumes := nil;
    //TODO implement

    pop_trace();
end;

function get_volume_list() : PDList;
var
    newList : PDList;
    i : uint32;
begin
    push_trace('VolumeManager.get_volume_list');

    //get copy of volume list
    newList := DL_New(sizeof(PStorage_Volume)); //TODO need to add DL_Copy function

    for i := 0 to DL_Size(volumes) - 1 do begin
        DL_Add(newList);
        DL_Set(newList, i, DL_Get(volumes, i));
    end;
end;

end.